#include "microcore.h"

#ifdef _WIN32
#include <windows.h>
#else
#include <time.h>
#include <sys/time.h>
#include <pthread.h>
#include <unistd.h>
#endif

// Mutex

Mutex::Mutex() : handle(0)
{
#ifdef _WIN32
    handle = new CRITICAL_SECTION;
    InitializeCriticalSection( (LPCRITICAL_SECTION)handle );
#else
    handle = new pthread_mutex_t;
    pthread_mutex_init( (pthread_mutex_t *)handle, 0 );
#endif
}

Mutex::~Mutex()
{
#ifdef _WIN32
    DeleteCriticalSection( (LPCRITICAL_SECTION)handle );
    delete (CRITICAL_SECTION *)handle;
#else
    pthread_mutex_destroy( (pthread_mutex_t *)handle );
    delete (pthread_mutex_t *)handle;
#endif
}

void Mutex::lock()
{
#ifdef _WIN32
    EnterCriticalSection( (LPCRITICAL_SECTION)handle );
#else
    pthread_mutex_lock( (pthread_mutex_t *)handle );
#endif
}

void Mutex::unlock()
{
#ifdef _WIN32
    LeaveCriticalSection( (LPCRITICAL_SECTION)handle );
#else
    pthread_mutex_unlock( (pthread_mutex_t *)handle );
#endif
}

// MutexLock

MutexLock::MutexLock( Mutex &m ) : mref( &m )
{
    mref->lock();
}

MutexLock::~MutexLock()
{
    mref->unlock();
}

// Event

// TODO: unix versions
Event::Event( bool autoReset )
{
    sys = CreateEvent( 0, autoReset ? FALSE : TRUE, FALSE, 0);
}

Event::~Event()
{
    if ( sys != 0 )
        CloseHandle( (HANDLE)sys );
}

// signal(set) event
void Event::signal()
{
    if ( sys != 0 )
        SetEvent( (HANDLE)sys );
}

// reset event
void Event::reset()
{
    if ( sys != 0 )
        ResetEvent( (HANDLE)sys );
}

// wait for this event
bool Event::wait( int ms )
{
    if ( sys != 0 )
    {
        return WaitForSingleObject( (HANDLE)sys,
                                    ms == -1 ? INFINITE : ms ) == WAIT_OBJECT_0;
    }
    return false;
}

// Thread

#ifdef _WIN32
static DWORD threadProc( LPVOID param )
{
    return ((Thread *)param)->run();
}
#else
static void *threadProc( void *param )
{
    return (void *)((Thread *)param)->run();
}
#endif

Thread::Thread( void *param ) : handle(0), thrParam( param ), shouldTerminate(0)
{
}

Thread::~Thread()
{
    shouldTerminate = 1;
#ifdef _WIN32
    if ( WaitForSingleObject( (HANDLE)handle, 500 ) != WAIT_OBJECT_0 )
    {
        TerminateThread( (HANDLE)handle, (DWORD)-1 );
    }
    CloseHandle( (HANDLE)handle );
#else
    pthread_setcanceltype( PTHREAD_CANCEL_ASYNCHRONOUS, 0 );
    pthread_cancel( *(pthread_t *)handle );
    delete (pthread_t *)handle;
#endif
}

// resume thread
bool Thread::resume()
{
    if ( !handle )
    {
#ifdef _WIN32
        handle = CreateThread( 0, 0, (LPTHREAD_START_ROUTINE)threadProc, (LPVOID)this, 0/*CREATE_SUSPENDED*/, 0 );
#else
        handle = new pthread_t;
        pthread_attr_t attr;
        pthread_attr_init(&attr);
        pthread_create( (pthread_t *)handle, 0 /* attr */, threadProc, this );
        pthread_attr_destroy(&attr);
#endif

        return handle != 0;
    }
#ifdef _WIN32
    return ResumeThread( (HANDLE)handle ) != (DWORD)-1;
#else
    // FIXME: pthreads don't have a simple mechanism for resuming a thread I think
    return 0;
#endif
}

// suspend thread
bool Thread::suspend()
{
#ifdef _WIN32
    return SuspendThread( (HANDLE)handle ) != (DWORD)-1;
#else
    // FIXME: pthreads don't have a simple mechanism for suspending a thread I think
    return 0;
#endif
}

// main worker proc
int Thread::run()
{
    return 0;
}

// sleep in ms
void Thread::sleep( int ms )
{
#ifdef _WIN32
    Sleep( ms == -1 ? INFINITE : ms );
#else
    ::usleep( ms * 1000 );
#endif
}

bool Thread::wait( Event &ev, int ms )
{
    return ev.wait( ms );
}

// statics

static Mutex counterMutex;
static int basesec;
static int counterInit = 0;

int getMillisec()
{
#ifndef _WIN32
    MutexLock _( counterMutex );
    struct timeval tp;
    struct timezone tzp;

    gettimeofday( &tp, &tzp );

    if ( !counterInit )
    {
        counterInit = 1;
        basesec = tp.tv_sec;
        return tp.tv_usec / 1000;
    }

    return (tp.tv_sec - basesec) * 1000 + tp.tv_usec / 1000;
#else
    return (int)GetTickCount();
#endif
}
